from typing import Dict, List, Optional, Set
from collections import namedtuple
import pandas as pd
from requests import Session
import json
from functools import lru_cache

__all__ = [
    "init",
    "perform_inference",
    "run_ner",
    "tag",
    "normalize",
    "find_abbreviations",
    "assign_control_pert_labels",
    "create_cohorts",
]

client = None


def init(token: str):
    global client
    client = _Client(token)


def _get_client() -> "_Client":
    global client
    if client is None:
        raise Exception("client has not been initialized with authentication token")
    return client


def perform_inference(model_name: str, input_data: dict) -> dict:
    """
    This is a wrapper around model inference APIs
    It serializes input_data, calls the API for the given model_name and returns deserialized output

    Args:
        model_name (str): one of 'normalizer', 'biobert' and 'control-perturbation'
        input_data (dict): model input

    Returns:
        dict
    """
    client = _get_client()
    url = client.base_url + model_name

    payload = json.dumps({"data": {"attributes": input_data, "type": "curation"}})
    response = client.session.post(url, data=payload)

    if response.status_code != 201:
        print("Response: ", response.text)
        raise Exception(f"Server responded with {response.status_code}")

    try:
        response = response.json()
    except json.JSONDecodeError as e:
        print("response: ", response.text)
        raise e

    if "data" in response:
        return response["data"]

    return response


def run_ner(
    text: str, threshold: Optional[float] = None, normalize_output: bool = False
):
    """Run an NER model on the given text. The returned value is a list of entities along with span info.

    Args:
        text (str): input text

    Kwargs:
        normalize_output (bool): whether to normalize the keywords

    Returns:
        entities (List[dict]): returns a list of spans, each span contains the keyword, start and end index of the keyword and the entity type
    """
    # TODO: If text is too long, break it up into chunks small enough for biobert

    # TODO: Why is this happening?
    text = text.replace("\n", "")
    payload = {"text": text}
    if threshold:
        payload["threshold"] = threshold
    response = perform_inference("biobert", payload)
    try:
        entities = response["entities"]
    except KeyError as e:
        print(response)
        raise e

    if normalize_output:
        for entity in entities:
            norm = normalize(entity["keyword"], entity["entity_type"], text)
            if norm["ontology"] == "CUI-less":
                entity["name"] = None
            else:
                entity["ontology_id"] = norm["ontology"] + ":" + norm["ontology_id"]
                entity["name"] = norm["name"]

    return entities


Tag = namedtuple("Tag", ["name", "ontology_id", "entity_type"])


def tag(text: str) -> List[Tag]:
    """
    Tag a given piece of text. A "tag" is just an ontology term.

    This function calls `run_ner` followed by `normalize`

    Args:
        text (str): Input text

    Returns:
        tags (set of tuples): set of unique tags
    """
    # TODO: remove non specific

    entities = run_ner(text, normalize_output=True)
    res = {
        Tag(e["name"], e["ontology_id"], e["entity_type"])
        for e in entities
        if e["name"]
    }
    return list(res)


def find_abbreviations(text: str) -> Dict[str, str]:
    data = {
        "mention": {"keyword": "dummykeyword", "entity_type": "gene"},
        "context": text,
    }

    output = perform_inference("normalizer", data)

    return output["abbreviations"]


@lru_cache(maxsize=None)
def normalize(
    mention: str,
    entity_type: str,
    context: Optional[str] = None,
    threshold: Optional[float] = None,
) -> dict:
    """Map a given mention (keyword) to an ontology term.

    Args:
        mention (str): mention of an entity e.g. "Cadiac arrythmia"
        entity_type (str): Should be one of ['disease', 'drug', 'tissue', 'cell_type', 'cell_line', 'species', 'gene']

    Kwargs:
        context (Optional[str]): The text where the mention occurs. This is used to resolve abbreviations

    Returns:
        dict

    """
    # TODO: Composite mentions?
    data = {
        "mention": {
            "keyword": mention,
            "entity_type": entity_type,
            "threshold": threshold,
        }
    }

    if context:
        data["context"] = context

    output = perform_inference("normalizer", data)

    if "term" not in output:
        return {
            "ontology": "CUI-less",
            "ontology_id": None,
            "name": None,
            "entity_type": entity_type,
        }

    return output["term"]


def assign_control_pert_labels(
    sample_metadata, columns_to_exclude=None
) -> pd.DataFrame:
    """Returns the sample metadata dataframe with 2 additional columns

    `is_control` - whether the sample is a control sample
    `control_prob` - the probability that the sample is control

    Args:
        sample_metadata (pd.DataFrame):

    Kwargs:
        columns_to_exclude (Set[str]): Any columns which don't play any role in determining the label, e.g. any arbitrary sample identifier

    Returns:
        pd.DataFrame

    """
    sample_metadata = sample_metadata.copy()

    if columns_to_exclude is None:
        columns_to_exclude = []

    cols = sample_metadata.columns.difference(columns_to_exclude)

    samples = sample_metadata[cols].to_dict("records")
    request_body = {"samples": samples}
    response = perform_inference("control-pertubation", request_body)

    response = {k: v for k, v in response.items() if k != "version" and k != "testing"}
    output = pd.DataFrame(response)
    sample_metadata["is_control"] = output["is_control"].values
    sample_metadata["control_prob"] = output["control_prob"].values
    return sample_metadata


def create_cohorts(
    sample_metadata: pd.DataFrame,
    columns_to_exclude: Optional[List[str]] = None,
    max_unique: int = 10,
) -> pd.DataFrame:
    """TODO: Docstring for assign_cohorts.

    Args:
        sample_metadata (pd.DataFrame): TODO

    Kwargs:
        columns_to_exclude (Set[str]): TODO
        max_unique (int): TODO

    Returns: TODO

    """
    # TODO
    raise NotImplementedError()


class _Client:
    """
    Stores authentication token
    """

    def __init__(self, token: str) -> None:
        """
        Args:
            token (str): polly user token
        """
        self.session = Session()
        self.session.headers = {
            "Content-Type": "application/vnd.api+json",
            "Cookie": f"refreshToken={token}",
        }
        self.base_url = "https://api.discover.polly.elucidata.io/curations/inferences/"
